=== modified file 'loggerhead/branchview.py'
--- loggerhead/branchview.py	2007-03-06 04:27:05 +0000
+++ loggerhead/branchview.py	2007-08-23 15:03:26 +0000
@@ -29,7 +29,7 @@
 from cherrypy import HTTPRedirect
 
 from loggerhead import util
-from loggerhead.changecache import ChangeCache
+from loggerhead.changecache import ChangeCache, FileChangeCache
 from loggerhead.history import History
 from loggerhead.textindex import TextIndex
 from loggerhead.controllers.changelog_ui import ChangeLogUI
@@ -141,14 +141,16 @@
             self.log.debug('Reload branch history...')
             if self._history is not None:
                 self._history.detach()
-            self._history = History.from_folder(self._absfolder, self._name)
+            _history = self._history = History.from_folder(
+                self._absfolder, self._name)
             cache_path = self._config.get('cachepath', None)
             if cache_path is None:
                 # try the project config
                 cache_path = self._project_config.get('cachepath', None)
             if cache_path is not None:
-                self._history.use_cache(ChangeCache(self._history, cache_path))
-                self._history.use_search_index(TextIndex(self._history, cache_path))
+                _history.use_cache(ChangeCache(_history, cache_path))
+                _history.use_file_cache(FileChangeCache(_history, cache_path))
+                _history.use_search_index(TextIndex(_history, cache_path))
         return self._history
     
     def check_rebuild(self):

=== modified file 'loggerhead/changecache.py'
--- loggerhead/changecache.py	2007-06-25 01:40:37 +0000
+++ loggerhead/changecache.py	2007-09-03 11:16:33 +0000
@@ -163,3 +163,38 @@
         self.flush()
 
 
+class FileChangeCache(object):
+    def __init__(self, history, cache_path):
+        self.history = history
+
+        if not os.path.exists(cache_path):
+            os.mkdir(cache_path)
+
+        self._changes_filename = os.path.join(cache_path, 'filechanges')
+
+        # use a lockfile since the cache folder could be shared across
+        # different processes.
+        self._lock = LockFile(os.path.join(cache_path, 'filechange-lock'))
+
+    @with_lock
+    def get_file_changes(self, entries):
+        cache = shelve.open(self._changes_filename, 'c', protocol=2)
+        try:
+            out = []
+            missing_entries = []
+            missing_entry_indices = []
+            for entry in entries:
+                changes = cache.get(entry.revid)
+                if changes is not None:
+                    out.append(changes)
+                else:
+                    missing_entries.append(entry)
+                    missing_entry_indices.append(len(out))
+                    out.append(None)
+            missing_changes = self.history.get_file_changes_uncached(missing_entries)
+            for i, entry, changes in zip(
+                missing_entry_indices, missing_entries, missing_changes):
+                cache[entry.revid] = out[i] = changes
+            return out
+        finally:
+            cache.close()

=== modified file 'loggerhead/controllers/annotate_ui.py'
--- loggerhead/controllers/annotate_ui.py	2007-06-19 13:49:09 +0000
+++ loggerhead/controllers/annotate_ui.py	2007-08-23 12:18:18 +0000
@@ -52,44 +52,48 @@
         z = time.time()
         h = self._branch.get_history()
         util.set_context(kw)
-        
-        if len(args) > 0:
-            revid = h.fix_revid(args[0])
-        else:
-            revid = h.last_revid
-
-        path = None
-        if len(args) > 1:
-            path = '/'.join(args[1:])
-            if not path.startswith('/'):
-                path = '/' + path
-        
-        file_id = kw.get('file_id', None)
-        if (file_id is None) and (path is None):
-            raise HTTPError(400, 'No file_id or filename provided to annotate')
-
-        if file_id is None:
-            file_id = h.get_file_id(revid, path)
-            
-        # no navbar for revisions
-        navigation = util.Container()
-        
-        if path is None:
-            path = h.get_path(revid, file_id)
-        filename = os.path.basename(path)
-
-        vals = {
-            'branch': self._branch,
-            'util': util,
-            'revid': revid,
-            'file_id': file_id,
-            'path': path,
-            'filename': filename,
-            'history': h,
-            'navigation': navigation,
-            'change': h.get_changes([ revid ])[0],
-            'contents': list(h.annotate_file(file_id, revid)),
-        }
-        h.flush_cache()
-        self.log.info('/annotate: %r secs' % (time.time() - z,))
-        return vals
+
+        h._branch.lock_read()
+        try:
+            if len(args) > 0:
+                revid = h.fix_revid(args[0])
+            else:
+                revid = h.last_revid
+
+            path = None
+            if len(args) > 1:
+                path = '/'.join(args[1:])
+                if not path.startswith('/'):
+                    path = '/' + path
+
+            file_id = kw.get('file_id', None)
+            if (file_id is None) and (path is None):
+                raise HTTPError(400, 'No file_id or filename provided to annotate')
+
+            if file_id is None:
+                file_id = h.get_file_id(revid, path)
+
+            # no navbar for revisions
+            navigation = util.Container()
+
+            if path is None:
+                path = h.get_path(revid, file_id)
+            filename = os.path.basename(path)
+
+            vals = {
+                'branch': self._branch,
+                'util': util,
+                'revid': revid,
+                'file_id': file_id,
+                'path': path,
+                'filename': filename,
+                'history': h,
+                'navigation': navigation,
+                'change': h.get_changes([ revid ])[0],
+                'contents': list(h.annotate_file(file_id, revid)),
+            }
+            h.flush_cache()
+            self.log.info('/annotate: %r secs' % (time.time() - z,))
+            return vals
+        finally:
+            h._branch.unlock()

=== modified file 'loggerhead/controllers/atom_ui.py'
--- loggerhead/controllers/atom_ui.py	2007-06-27 17:06:57 +0000
+++ loggerhead/controllers/atom_ui.py	2007-08-30 11:39:10 +0000
@@ -33,17 +33,22 @@
     @turbogears.expose(template='loggerhead.templates.atom', format="xml", content_type="application/atom+xml")
     def default(self, *args):
         h = self._branch.get_history()
-        pagesize = int(self._branch.config.get('pagesize', '20'))
-
-        revid_list = h.get_file_view(h.last_revid, None)
-        entries = list(h.get_changes(list(revid_list)[:pagesize]))
-
-        vals = {
-            'branch': self._branch,
-            'changes': entries,
-            'util': util,
-            'history': h,
-            'updated': entries[0].date.isoformat() + 'Z',
-        }
-        h.flush_cache()
-        return vals
+
+        h._branch.lock_read()
+        try:
+            pagesize = int(self._branch.config.get('pagesize', '20'))
+
+            revid_list = h.get_file_view(h.last_revid, None)
+            entries = list(h.get_changes(list(revid_list)[:pagesize]))
+
+            vals = {
+                'branch': self._branch,
+                'changes': entries,
+                'util': util,
+                'history': h,
+                'updated': entries[0].date.isoformat() + 'Z',
+            }
+            h.flush_cache()
+            return vals
+        finally:
+            h._branch.unlock()

=== modified file 'loggerhead/controllers/bundle_ui.py'
--- loggerhead/controllers/bundle_ui.py	2007-06-26 04:32:43 +0000
+++ loggerhead/controllers/bundle_ui.py	2007-08-23 12:18:18 +0000
@@ -38,22 +38,26 @@
         z = time.time()
         h = self._branch.get_history()
 
-        if len(args) < 1:
-            raise HTTPRedirect(self._branch.url('/changes'))
-        revid = h.fix_revid(args[0])
-        if len(args) >= 3:
-            compare_revid = h.fix_revid(args[1])
-        else:
-            compare_revid = None
-        
+        h._branch.lock_read()
         try:
-            bundle_data = h.get_bundle(revid, compare_revid)
-        except:
-            self.log.exception('Exception fetching bundle')
-            raise InternalError('Could not fetch bundle')
-            
-        response.headers['Content-Type'] = 'text/plain'
-        response.headers['Content-Length'] = len(bundle_data)
-        response.body = bundle_data
-        self.log.info('/bundle: %r seconds' % (time.time() - z,))
-        return response.body
+            if len(args) < 1:
+                raise HTTPRedirect(self._branch.url('/changes'))
+            revid = h.fix_revid(args[0])
+            if len(args) >= 3:
+                compare_revid = h.fix_revid(args[1])
+            else:
+                compare_revid = None
+
+            try:
+                bundle_data = h.get_bundle(revid, compare_revid)
+            except:
+                self.log.exception('Exception fetching bundle')
+                raise InternalError('Could not fetch bundle')
+
+            response.headers['Content-Type'] = 'text/plain'
+            response.headers['Content-Length'] = len(bundle_data)
+            response.body = bundle_data
+            self.log.info('/bundle: %r seconds' % (time.time() - z,))
+            return response.body
+        finally:
+            h._branch.unlock()

=== modified file 'loggerhead/controllers/changelog_ui.py'
--- loggerhead/controllers/changelog_ui.py	2007-07-04 12:05:56 +0000
+++ loggerhead/controllers/changelog_ui.py	2007-08-23 12:18:18 +0000
@@ -39,77 +39,82 @@
         z = time.time()
         h = self._branch.get_history()
         config = self._branch.config
-        
-        if len(args) > 0:
-            revid = h.fix_revid(args[0])
-        else:
-            revid = None
 
-        file_id = kw.get('file_id', None)
-        query = kw.get('q', None)
-        start_revid = h.fix_revid(kw.get('start_revid', None))
-        orig_start_revid = start_revid
-        pagesize = int(config.get('pagesize', '20'))
-        search_failed = False
-        
+        h._branch.lock_read()
         try:
-            revid, start_revid, revid_list = h.get_view(revid, start_revid, file_id, query)
-            kw['start_revid'] = start_revid
-            util.set_context(kw)
-            
-            if (query is not None) and (len(revid_list) == 0):
-                search_failed = True
-
-            if len(revid_list) == 0:
-                scan_list = revid_list
+            if len(args) > 0:
+                revid = h.fix_revid(args[0])
             else:
-                if revid in revid_list: # XXX is this always true?
-                    i = revid_list.index(revid)
+                revid = None
+
+            file_id = kw.get('file_id', None)
+            query = kw.get('q', None)
+            start_revid = h.fix_revid(kw.get('start_revid', None))
+            orig_start_revid = start_revid
+            pagesize = int(config.get('pagesize', '20'))
+            search_failed = False
+
+            try:
+                revid, start_revid, revid_list = h.get_view(revid, start_revid, file_id, query)
+                kw['start_revid'] = start_revid
+                util.set_context(kw)
+
+                if (query is not None) and (len(revid_list) == 0):
+                    search_failed = True
+
+                if len(revid_list) == 0:
+                    scan_list = revid_list
                 else:
-                    i = None
-                scan_list = revid_list[i:]
-            entry_list = scan_list[:pagesize]
-            entries = h.get_changes(entry_list)
-        except:
-            self.log.exception('Exception fetching changes')
-            raise InternalError('Could not fetch changes')
-
-        navigation = util.Container(pagesize=pagesize, revid=revid, start_revid=start_revid, revid_list=revid_list,
-                                    file_id=file_id, scan_url='/changes', branch=self._branch, feed=True)
-        if query is not None:
-            navigation.query = query
-        util.fill_in_navigation(navigation)
-        
-        entries = list(entries)
-        # add parent & merge-point branch-nick info, in case it's useful
-        h.get_branch_nicks(entries)
-        
-        # does every change on this page have the same committer?  if so,
-        # tell the template to show committer info in the "details block"
-        # instead of on each line.
-        all_same_author = True
-
-        if entries:
-            author = entries[0].author
-            for e in entries[1:]:
-                if e.author != author:
-                    all_same_author = False
-                    break
-
-        vals = {
-            'branch': self._branch,
-            'changes': entries,
-            'util': util,
-            'history': h,
-            'revid': revid,
-            'navigation': navigation,
-            'file_id': file_id,
-            'start_revid': start_revid,
-            'viewing_from': (orig_start_revid is not None) and (orig_start_revid != h.last_revid),
-            'query': query,
-            'search_failed': search_failed,
-            'all_same_author': all_same_author,
-        }
-        h.flush_cache()
-        self.log.info('/changes %r: %r secs' % (revid, time.time() - z))
-        return vals
+                    if revid in revid_list: # XXX is this always true?
+                        i = revid_list.index(revid)
+                    else:
+                        i = None
+                    scan_list = revid_list[i:]
+                entry_list = scan_list[:pagesize]
+                entries = h.get_changes(entry_list)
+                h.add_changes(entries)
+            except:
+                self.log.exception('Exception fetching changes')
+                raise InternalError('Could not fetch changes')
+
+            navigation = util.Container(pagesize=pagesize, revid=revid, start_revid=start_revid, revid_list=revid_list,
+                                        file_id=file_id, scan_url='/changes', branch=self._branch, feed=True)
+            if query is not None:
+                navigation.query = query
+            util.fill_in_navigation(navigation)
+
+            entries = list(entries)
+            # add parent & merge-point branch-nick info, in case it's useful
+            h.get_branch_nicks(entries)
+
+            # does every change on this page have the same committer?  if so,
+            # tell the template to show committer info in the "details block"
+            # instead of on each line.
+            all_same_author = True
+
+            if entries:
+                author = entries[0].author
+                for e in entries[1:]:
+                    if e.author != author:
+                        all_same_author = False
+                        break
+
+            vals = {
+                'branch': self._branch,
+                'changes': entries,
+                'util': util,
+                'history': h,
+                'revid': revid,
+                'navigation': navigation,
+                'file_id': file_id,
+                'start_revid': start_revid,
+                'viewing_from': (orig_start_revid is not None) and (orig_start_revid != h.last_revid),
+                'query': query,
+                'search_failed': search_failed,
+                'all_same_author': all_same_author,
+            }
+            h.flush_cache()
+            self.log.info('/changes %r: %r secs' % (revid, time.time() - z))
+            return vals
+        finally:
+            h._branch.unlock()

=== modified file 'loggerhead/controllers/download_ui.py'
--- loggerhead/controllers/download_ui.py	2007-05-23 17:42:19 +0000
+++ loggerhead/controllers/download_ui.py	2007-08-23 12:18:18 +0000
@@ -47,19 +47,23 @@
         z = time.time()
         h = self._branch.get_history()
         
-        if len(args) < 2:
-            raise HTTPRedirect(self._branch.url('/changes'))
-        
-        revid = h.fix_revid(args[0])
-        file_id = args[1]
-        path, filename, content = h.get_file(file_id, revid)
-        mime_type, encoding = mimetypes.guess_type(filename)
-        if mime_type is None:
-            mime_type = 'application/octet-stream'
-
-        self.log.info('/download %s @ %s (%d bytes)', path, h.get_revno(revid), len(content))
-        response.headers['Content-Type'] = mime_type
-        response.headers['Content-Length'] = len(content)
-        response.headers['Content-Disposition'] = 'attachment; filename=%s'%(filename,)
-        response.body = content
-        return response.body
+        h._branch.lock_read()
+        try:
+            if len(args) < 2:
+                raise HTTPRedirect(self._branch.url('/changes'))
+
+            revid = h.fix_revid(args[0])
+            file_id = args[1]
+            path, filename, content = h.get_file(file_id, revid)
+            mime_type, encoding = mimetypes.guess_type(filename)
+            if mime_type is None:
+                mime_type = 'application/octet-stream'
+
+            self.log.info('/download %s @ %s (%d bytes)', path, h.get_revno(revid), len(content))
+            response.headers['Content-Type'] = mime_type
+            response.headers['Content-Length'] = len(content)
+            response.headers['Content-Disposition'] = 'attachment; filename=%s'%(filename,)
+            response.body = content
+            return response.body
+        finally:
+            h._branch.unlock()

=== modified file 'loggerhead/controllers/inventory_ui.py'
--- loggerhead/controllers/inventory_ui.py	2007-06-27 16:52:56 +0000
+++ loggerhead/controllers/inventory_ui.py	2007-08-22 22:39:43 +0000
@@ -52,63 +52,59 @@
         z = time.time()
         h = self._branch.get_history()
         util.set_context(kw)
-        
-        if len(args) > 0:
-            revid = h.fix_revid(args[0])
-        else:
-            revid = h.last_revid
-        
-        try:
-            rev = h.get_revision(revid)
-            inv = h.get_inventory(revid)
-        except:
-            self.log.exception('Exception fetching changes')
-            raise InternalError('Could not fetch changes')
-
-        file_id = kw.get('file_id', inv.root.file_id)
-        sort_type = kw.get('sort', None)
-
-        try:
-            rev = h.get_revision(revid)
-            inv = h.get_inventory(revid)
-        except:
-            self.log.exception('Exception fetching changes')
-            raise InternalError('Could not fetch changes')
-
-        # no navbar for revisions
-        navigation = util.Container()
-
-        change = h.get_changes([ revid ])[0]
-        # add parent & merge-point branch-nick info, in case it's useful
-        h.get_branch_nicks([ change ])
-        
-        path = inv.id2path(file_id)
-        if not path.startswith('/'):
-            path = '/' + path
-        idpath = inv.get_idpath(file_id)
-        if len(idpath) > 1:
-            updir = dirname(path)
-            updir_file_id = idpath[-2]
-        else:
-            updir = None
-            updir_file_id = None
-        if updir == '/':
-            updir_file_id = None
-
-        vals = {
-            'branch': self._branch,
-            'util': util,
-            'revid': revid,
-            'change': change,
-            'file_id': file_id,
-            'path': path,
-            'updir': updir,
-            'updir_file_id': updir_file_id,
-            'filelist': h.get_filelist(inv, file_id, sort_type),
-            'history': h,
-            'posixpath': posixpath,
-            'navigation': navigation,
-        }
-        h.flush_cache()
-        self.log.info('/inventory %r: %r secs' % (revid, time.time() - z))
-        return vals
+
+        h._branch.lock_read()
+        try:
+            if len(args) > 0:
+                revid = h.fix_revid(args[0])
+            else:
+                revid = h.last_revid
+
+            try:
+                inv = h.get_inventory(revid)
+            except:
+                self.log.exception('Exception fetching changes')
+                raise InternalError('Could not fetch changes')
+
+            file_id = kw.get('file_id', inv.root.file_id)
+            sort_type = kw.get('sort', None)
+
+            # no navbar for revisions
+            navigation = util.Container()
+
+            change = h.get_changes([ revid ])[0]
+            # add parent & merge-point branch-nick info, in case it's useful
+            h.get_branch_nicks([ change ])
+
+            path = inv.id2path(file_id)
+            if not path.startswith('/'):
+                path = '/' + path
+            idpath = inv.get_idpath(file_id)
+            if len(idpath) > 1:
+                updir = dirname(path)
+                updir_file_id = idpath[-2]
+            else:
+                updir = None
+                updir_file_id = None
+            if updir == '/':
+                updir_file_id = None
+
+            vals = {
+                'branch': self._branch,
+                'util': util,
+                'revid': revid,
+                'change': change,
+                'file_id': file_id,
+                'path': path,
+                'updir': updir,
+                'updir_file_id': updir_file_id,
+                'filelist': h.get_filelist(inv, file_id, sort_type),
+                'history': h,
+                'posixpath': posixpath,
+                'navigation': navigation,
+            }
+            h.flush_cache()
+            self.log.info('/inventory %r: %r secs' % (revid, time.time() - z))
+            return vals
+        finally:
+            h._branch.unlock()

=== modified file 'loggerhead/controllers/revision_ui.py'
--- loggerhead/controllers/revision_ui.py	2007-06-12 15:57:50 +0000
+++ loggerhead/controllers/revision_ui.py	2007-08-23 12:18:18 +0000
@@ -44,55 +44,56 @@
         h = self._branch.get_history()
         util.set_context(kw)
         
-        if len(args) > 0:
-            revid = h.fix_revid(args[0])
-        else:
-            revid = None
-        
-        file_id = kw.get('file_id', None)
-        start_revid = h.fix_revid(kw.get('start_revid', None))
-        query = kw.get('q', None)
-        remember = kw.get('remember', None)
-        compare_revid = kw.get('compare_revid', None)
-        
+        h._branch.lock_read()
         try:
-            revid, start_revid, revid_list = h.get_view(revid, start_revid, file_id, query)
-        except:
-            self.log.exception('Exception fetching changes')
-            raise InternalError('Could not fetch changes')
-        
-        navigation = util.Container(revid_list=revid_list, revid=revid, start_revid=start_revid, file_id=file_id,
-                                    pagesize=1, scan_url='/revision', branch=self._branch, feed=True)
-        if query is not None:
-            navigation.query = query
-        util.fill_in_navigation(navigation)
-
-        if compare_revid is not None:
-            change = h.get_diff(compare_revid, revid)
-        else:
-            change = h.get_changes([ revid ], get_diffs=True)[0]
-        # add parent & merge-point branch-nick info, in case it's useful
-        h.get_branch_nicks([ change ])
-
-        # let's make side-by-side diff be the default
-        side_by_side = not kw.get('unified', False)
-        if side_by_side:
-            h.add_side_by_side([ change ])
-        
-        vals = {
-            'branch': self._branch,
-            'revid': revid,
-            'change': change,
-            'start_revid': start_revid,
-            'file_id': file_id,
-            'util': util,
-            'history': h,
-            'navigation': navigation,
-            'query': query,
-            'remember': remember,
-            'compare_revid': compare_revid,
-            'side_by_side': side_by_side,
-        }
-        h.flush_cache()
-        self.log.info('/revision: %r seconds' % (time.time() - z,))
-        return vals
+            if len(args) > 0:
+                revid = h.fix_revid(args[0])
+            else:
+                revid = None
+
+            file_id = kw.get('file_id', None)
+            start_revid = h.fix_revid(kw.get('start_revid', None))
+            query = kw.get('q', None)
+            remember = kw.get('remember', None)
+            compare_revid = kw.get('compare_revid', None)
+
+            try:
+                revid, start_revid, revid_list = h.get_view(revid, start_revid, file_id, query)
+            except:
+                self.log.exception('Exception fetching changes')
+                raise InternalError('Could not fetch changes')
+
+            navigation = util.Container(revid_list=revid_list, revid=revid, start_revid=start_revid, file_id=file_id,
+                                        pagesize=1, scan_url='/revision', branch=self._branch, feed=True)
+            if query is not None:
+                navigation.query = query
+            util.fill_in_navigation(navigation)
+
+            change = h.get_change_with_diff(revid, compare_revid)
+            # add parent & merge-point branch-nick info, in case it's useful
+            h.get_branch_nicks([ change ])
+
+            # let's make side-by-side diff be the default
+            side_by_side = not kw.get('unified', False)
+            if side_by_side:
+                h.add_side_by_side([ change ])
+
+            vals = {
+                'branch': self._branch,
+                'revid': revid,
+                'change': change,
+                'start_revid': start_revid,
+                'file_id': file_id,
+                'util': util,
+                'history': h,
+                'navigation': navigation,
+                'query': query,
+                'remember': remember,
+                'compare_revid': compare_revid,
+                'side_by_side': side_by_side,
+            }
+            h.flush_cache()
+            self.log.info('/revision: %r seconds' % (time.time() - z,))
+            return vals
+        finally:
+            h._branch.unlock()

=== modified file 'loggerhead/history.py'
--- loggerhead/history.py	2007-06-27 17:06:57 +0000
+++ loggerhead/history.py	2007-09-07 08:09:52 +0000
@@ -51,6 +51,7 @@
 import bzrlib.diff
 import bzrlib.errors
 import bzrlib.progress
+import bzrlib.revision
 import bzrlib.textfile
 import bzrlib.tsort
 import bzrlib.ui
@@ -58,18 +59,6 @@
 
 with_branch_lock = util.with_lock('_lock', 'branch')
 
-@decorator
-def with_bzrlib_read_lock(unbound):
-    def bzrlib_read_locked(self, *args, **kw):
-        #self.log.debug('-> %r bzr lock', id(threading.currentThread()))
-        self._branch.repository.lock_read()
-        try:
-            return unbound(self, *args, **kw)
-        finally:
-            self._branch.repository.unlock()
-            #self.log.debug('<- %r bzr lock', id(threading.currentThread()))
-    return bzrlib_read_locked
-
 
 # bzrlib's UIFactory is not thread-safe
 uihack = threading.local()
@@ -100,7 +89,7 @@
     """
     turn a normal unified-style diff (post-processed by parse_delta) into a
     side-by-side diff structure.  the new structure is::
-    
+
         chunks: list(
             diff: list(
                 old_lineno: int,
@@ -150,12 +139,12 @@
     elif len(message) == 0:
         # sometimes a commit may have NO message!
         message = ['']
-        
+
     # make short form of commit message
     short_message = message[0]
     if len(short_message) > 80:
         short_message = short_message[:80] + '...'
-    
+
     return message, short_message
 
 
@@ -165,7 +154,7 @@
     if kind == 'symlink':
         path += '@'
     return path
-        
+
 
 
 # from bzrlib
@@ -187,12 +176,13 @@
 
 
 class History (object):
-    
+
     def __init__(self):
         self._change_cache = None
+        self._file_change_cache = None
         self._index = None
         self._lock = threading.RLock()
-    
+
     @classmethod
     def from_branch(cls, branch, name=None):
         z = time.time()
@@ -203,12 +193,12 @@
             self._revision_graph = branch.repository.get_revision_graph(self._last_revid)
         else:
             self._revision_graph = {}
-        
+
         if name is None:
             name = self._branch.nick
         self._name = name
         self.log = logging.getLogger('loggerhead.%s' % (name,))
-        
+
         self._full_history = []
         self._revision_info = {}
         self._revno_revid = {}
@@ -222,14 +212,14 @@
         # cache merge info
         self._where_merged = {}
         for revid in self._revision_graph.keys():
-            if not revid in self._full_history: 
+            if not revid in self._full_history:
                 continue
             for parent in self._revision_graph[revid]:
                 self._where_merged.setdefault(parent, set()).add(revid)
 
         self.log.info('built revision graph cache: %r secs' % (time.time() - z,))
         return self
-    
+
     @classmethod
     def from_folder(cls, path, name=None):
         b = bzrlib.branch.Branch.open(path)
@@ -245,7 +235,10 @@
 
     def use_cache(self, cache):
         self._change_cache = cache
-    
+
+    def use_file_cache(self, cache):
+        self._file_change_cache = cache
+
     def use_search_index(self, index):
         self._index = index
 
@@ -267,23 +260,19 @@
         if self._change_cache is None:
             return
         self._change_cache.flush()
-    
+
     def check_rebuild(self):
         if self._change_cache is not None:
             self._change_cache.check_rebuild()
         if self._index is not None:
             self._index.check_rebuild()
-    
+
     last_revid = property(lambda self: self._last_revid, None, None)
-    
+
     @with_branch_lock
     def get_config(self):
         return self._branch.get_config()
-    
-    @with_branch_lock
-    def get_revision(self, revid):
-        return self._branch.repository.get_revision(revid)
-    
+
     def get_revno(self, revid):
         if revid not in self._revision_info:
             # ghost parent?
@@ -293,7 +282,7 @@
 
     def get_revision_history(self):
         return self._full_history
-    
+
     def get_revids_from(self, revid_list, revid):
         """
         given a list of revision ids, yield revisions in graph order,
@@ -309,7 +298,7 @@
             if len(parents) == 0:
                 return
             revid = parents[0]
-    
+
     @with_branch_lock
     def get_short_revision_history_by_fileid(self, file_id):
         # wow.  is this really the only way we can get this list?  by
@@ -334,7 +323,7 @@
         revid_list.reverse()
         index = -index
         return revid_list[index:]
-    
+
     @with_branch_lock
     def get_revision_history_matching(self, revid_list, text):
         self.log.debug('searching %d revisions for %r', len(revid_list), text)
@@ -359,12 +348,12 @@
         # put them in some coherent order :)
         out = [r for r in self._full_history if r in out]
         return out
-    
+
     @with_branch_lock
     def get_search_revid_list(self, query, revid_list):
         """
         given a "quick-search" query, try a few obvious possible meanings:
-        
+
             - revision id or # ("128.1.3")
             - date (US style "mm/dd/yy", earth style "dd-mm-yy", or iso style "yyyy-mm-dd")
             - comment text as a fallback
@@ -374,14 +363,14 @@
         # FIXME: there is some silliness in this action.  we have to look up
         # all the relevant changes (time-consuming) only to return a list of
         # revids which will be used to fetch a set of changes again.
-        
+
         # if they entered a revid, just jump straight there; ignore the passed-in revid_list
         revid = self.fix_revid(query)
         if revid is not None:
             changes = self.get_changes([ revid ])
             if (changes is not None) and (len(changes) > 0):
                 return [ revid ]
-        
+
         date = None
         m = self.us_date_re.match(query)
         if m is not None:
@@ -399,12 +388,12 @@
                 # if no limit to the query was given, search only the direct-parent path.
                 revid_list = list(self.get_revids_from(None, self._last_revid))
             return self.get_revision_history_since(revid_list, date)
-        
+
         # check comment fields.
         if revid_list is None:
             revid_list = self._full_history
         return self.get_revision_history_matching_indexed(revid_list, query)
-    
+
     revno_re = re.compile(r'^[\d\.]+$')
     # the date regex are without a final '$' so that queries like
     # "2006-11-30 12:15" still mostly work.  (i think it's better to give
@@ -422,14 +411,14 @@
         if self.revno_re.match(revid):
             revid = self._revno_revid[revid]
         return revid
-    
+
     @with_branch_lock
     def get_file_view(self, revid, file_id):
         """
         Given a revid and optional path, return a (revlist, revid) for
         navigation through the current scope: from the revid (or the latest
         revision) back to the original revision.
-        
+
         If file_id is None, the entire revision history is the list scope.
         """
         if revid is None:
@@ -442,27 +431,27 @@
         else:
             revlist = list(self.get_revids_from(None, revid))
         return revlist
-    
+
     @with_branch_lock
     def get_view(self, revid, start_revid, file_id, query=None):
         """
         use the URL parameters (revid, start_revid, file_id, and query) to
         determine the revision list we're viewing (start_revid, file_id, query)
         and where we are in it (revid).
-        
+
         if a query is given, we're viewing query results.
         if a file_id is given, we're viewing revisions for a specific file.
         if a start_revid is given, we're viewing the branch from a
             specific revision up the tree.
         (these may be combined to view revisions for a specific file, from
             a specific revision, with a specific search query.)
-            
+
         returns a new (revid, start_revid, revid_list, scan_list) where:
-        
+
             - revid: current position within the view
             - start_revid: starting revision of this view
             - revid_list: list of revision ids for this view
-        
+
         file_id and query are never changed so aren't returned, but they may
         contain vital context for future url navigation.
         """
@@ -479,7 +468,7 @@
                 revid_list= self.get_file_view(revid, file_id)
                 start_revid = revid
             return revid, start_revid, revid_list
-        
+
         # potentially limit the search
         if file_id is not None:
             revid_list = self.get_file_view(start_revid, file_id)
@@ -507,13 +496,13 @@
         if (len(path) > 0) and not path.startswith('/'):
             path = '/' + path
         return path
-    
+
     @with_branch_lock
     def get_file_id(self, revid, path):
         if (len(path) > 0) and not path.startswith('/'):
             path = '/' + path
         return self._branch.repository.get_revision_inventory(revid).path2id(path)
-    
+
 
     def get_merge_point_list(self, revid):
         """
@@ -521,7 +510,7 @@
         """
         if '.' not in self.get_revno(revid):
             return []
-        
+
         merge_point = []
         while True:
             children = self._where_merged.get(revid, [])
@@ -544,7 +533,7 @@
                 merge_point.extend(merge_point_next)
 
             revid = nexts[0]
-            
+
     def simplify_merge_point_list(self, revids):
         """if a revision is already merged, don't show further merge points"""
         d = {}
@@ -587,35 +576,28 @@
                     p.branch_nick = p_change_dict[p.revid].branch_nick
                 else:
                     p.branch_nick = '(missing)'
-    
+
     @with_branch_lock
-    def get_changes(self, revid_list, get_diffs=False):
+    def get_changes(self, revid_list):
         if self._change_cache is None:
             changes = self.get_changes_uncached(revid_list)
         else:
             changes = self._change_cache.get_changes(revid_list)
         if len(changes) == 0:
             return changes
-        
+
         # some data needs to be recalculated each time, because it may
         # change as new revisions are added.
         for change in changes:
             merge_revids = self.simplify_merge_point_list(self.get_merge_point_list(change.revid))
             change.merge_points = [util.Container(revid=r, revno=self.get_revno(r)) for r in merge_revids]
             change.revno = self.get_revno(change.revid)
-            if get_diffs:
-                # this may be time-consuming, but it only happens on the revision page, and only for one revision at a time.
-                if len(change.parents) > 0:
-                    parent_revid = change.parents[0].revid
-                else:
-                    parent_revid = None
-                change.changes.modified = self._parse_diffs(parent_revid, change.revid)
 
         parity = 0
         for change in changes:
             change.parity = parity
             parity ^= 1
-        
+
         return changes
 
     # alright, let's profile this sucka.
@@ -629,36 +611,37 @@
         self.log.info('lsprof complete!')
         return ret
 
-    def _get_deltas_for_revisions_with_trees(self, revisions):
+    def _get_deltas_for_revisions_with_trees(self, entries):
         """Produce a generator of revision deltas.
-        
+
         Note that the input is a sequence of REVISIONS, not revision_ids.
         Trees will be held in memory until the generator exits.
         Each delta is relative to the revision's lefthand predecessor.
         """
         required_trees = set()
-        for revision in revisions:
-            required_trees.add(revision.revision_id)
-            required_trees.update(revision.parent_ids[:1])
-        trees = dict((t.get_revision_id(), t) for 
+        for entry in entries:
+            required_trees.add(entry.revid)
+            required_trees.update([p.revid for p in entry.parents[:1]])
+        trees = dict((t.get_revision_id(), t) for
                      t in self._branch.repository.revision_trees(required_trees))
         ret = []
         self._branch.repository.lock_read()
         try:
-            for revision in revisions:
-                if not revision.parent_ids:
-                    old_tree = self._branch.repository.revision_tree(None)
+            for entry in entries:
+                if not entry.parents:
+                    old_tree = self._branch.repository.revision_tree(
+                        bzrlib.revision.NULL_REVISION)
                 else:
-                    old_tree = trees[revision.parent_ids[0]]
-                tree = trees[revision.revision_id]
-                ret.append((tree, old_tree, tree.changes_from(old_tree)))
+                    old_tree = trees[entry.parents[0].revid]
+                tree = trees[entry.revid]
+                ret.append(tree.changes_from(old_tree))
             return ret
         finally:
             self._branch.repository.unlock()
-    
+
     def entry_from_revision(self, revision):
         commit_time = datetime.datetime.fromtimestamp(revision.timestamp)
-        
+
         parents = [util.Container(revid=r, revno=self.get_revno(r)) for r in revision.parent_ids]
 
         message, short_message = clean_message(revision.message)
@@ -676,45 +659,64 @@
         return util.Container(entry)
 
     @with_branch_lock
-    @with_bzrlib_read_lock
     def get_changes_uncached(self, revid_list):
-        done = False
-        while not done:
-            try:
-                rev_list = self._branch.repository.get_revisions(revid_list)
-                done = True
-            except (KeyError, bzrlib.errors.NoSuchRevision), e:
-                # this sometimes happens with arch-converted branches.
-                # i don't know why. :(
-                self.log.debug('No such revision (skipping): %s', e)
-                revid_list.remove(e.revision)
-        if not revid_list:
-            return []
-        
-        delta_list = self._get_deltas_for_revisions_with_trees(rev_list)
-        combined_list = zip(rev_list, delta_list)
-        
-        entries = []
-        for rev, (new_tree, old_tree, delta) in combined_list:
-            entry = self.entry_from_revision(rev)
-            entry.changes = self.parse_delta(delta, old_tree, new_tree)
-            entries.append(entry)
-        
-        return entries
-
-    @with_bzrlib_read_lock
-    def _get_diff(self, revid1, revid2):
-        rev_tree1 = self._branch.repository.revision_tree(revid1)
-        rev_tree2 = self._branch.repository.revision_tree(revid2)
+        # Because we may loop and call get_revisions multiple times (to throw
+        # out dud revids), we grab a read lock.
+        self._branch.lock_read()
+        try:
+            while True:
+                try:
+                    rev_list = self._branch.repository.get_revisions(revid_list)
+                except (KeyError, bzrlib.errors.NoSuchRevision), e:
+                    # this sometimes happens with arch-converted branches.
+                    # i don't know why. :(
+                    self.log.debug('No such revision (skipping): %s', e)
+                    revid_list.remove(e.revision)
+                else:
+                    break
+
+            return [self.entry_from_revision(rev) for rev in rev_list]
+        finally:
+            self._branch.unlock()
+
+    def get_file_changes_uncached(self, entries):
+        delta_list = self._get_deltas_for_revisions_with_trees(entries)
+
+        return [self.parse_delta(delta) for delta in delta_list]
+
+    @with_branch_lock
+    def get_file_changes(self, entries):
+        if self._file_change_cache is None:
+            return self.get_file_changes_uncached(entries)
+        else:
+            return self._file_change_cache.get_file_changes(entries)
+
+    def add_changes(self, entries):
+        changes_list = self.get_file_changes(entries)
+
+        for entry, changes in zip(entries, changes_list):
+            entry.changes = changes
+
+    @with_branch_lock
+    def get_change_with_diff(self, revid, compare_revid=None):
+        entry = self.get_changes([revid])[0]
+
+        if compare_revid is None:
+            if entry.parents:
+                compare_revid = entry.parents[0].revid
+            else:
+                compare_revid = 'null:'
+
+        rev_tree1 = self._branch.repository.revision_tree(compare_revid)
+        rev_tree2 = self._branch.repository.revision_tree(revid)
         delta = rev_tree2.changes_from(rev_tree1)
-        return rev_tree1, rev_tree2, delta
-    
-    def get_diff(self, revid1, revid2):
-        rev_tree1, rev_tree2, delta = self._get_diff(revid1, revid2)
-        entry = self.get_changes([ revid2 ], False)[0]
-        entry.changes = self.parse_delta(delta, True, rev_tree1, rev_tree2)
+
+        entry.changes = self.parse_delta(delta)
+
+        entry.changes.modified = self._parse_diffs(rev_tree1, rev_tree2, delta)
+
         return entry
-    
+
     @with_branch_lock
     def get_file(self, file_id, revid):
         "returns (path, filename, data)"
@@ -725,11 +727,11 @@
         if not path.startswith('/'):
             path = '/' + path
         return path, inv_entry.name, rev_tree.get_file_text(file_id)
-    
-    def _parse_diffs(self, revid1, revid2):
+
+    def _parse_diffs(self, old_tree, new_tree, delta):
         """
         Return a list of processed diffs, in the format::
-        
+
             list(
                 filename: str,
                 file_id: str,
@@ -743,16 +745,15 @@
                 ),
             )
         """
-        old_tree, new_tree, delta = self._get_diff(revid1, revid2)
         process = []
         out = []
-        
+
         for old_path, new_path, fid, kind, text_modified, meta_modified in delta.renamed:
             if text_modified:
                 process.append((old_path, new_path, fid, kind))
         for path, fid, kind, text_modified, meta_modified in delta.modified:
             process.append((path, path, fid, kind))
-        
+
         for old_path, new_path, fid, kind in process:
             old_lines = old_tree.get_file_lines(fid)
             new_lines = new_tree.get_file_lines(fid)
@@ -768,7 +769,7 @@
             else:
                 diff = ''
             out.append(util.Container(filename=rich_filename(new_path, kind), file_id=fid, chunks=self._process_diff(diff)))
-        
+
         return out
 
     def _process_diff(self, diff):
@@ -808,12 +809,11 @@
         if chunk is not None:
             chunks.append(chunk)
         return chunks
-                
-    @with_branch_lock
-    def parse_delta(self, delta, get_diffs=True, old_tree=None, new_tree=None):
+
+    def parse_delta(self, delta):
         """
         Return a nested data structure containing the changes in a delta::
-        
+
             added: list((filename, file_id)),
             renamed: list((old_filename, new_filename, file_id)),
             deleted: list((filename, file_id)),
@@ -826,21 +826,21 @@
         modified = []
         renamed = []
         removed = []
-        
+
         for path, fid, kind in delta.added:
             added.append((rich_filename(path, kind), fid))
-        
+
         for path, fid, kind, text_modified, meta_modified in delta.modified:
             modified.append(util.Container(filename=rich_filename(path, kind), file_id=fid))
-        
+
         for old_path, new_path, fid, kind, text_modified, meta_modified in delta.renamed:
             renamed.append((rich_filename(old_path, kind), rich_filename(new_path, kind), fid))
             if meta_modified or text_modified:
                 modified.append(util.Container(filename=rich_filename(new_path, kind), file_id=fid))
-        
+
         for path, fid, kind in delta.removed:
             removed.append((rich_filename(path, kind), fid))
-        
+
         return util.Container(added=added, renamed=renamed, removed=removed, modified=modified)
 
     @staticmethod
@@ -849,7 +849,7 @@
         for change in changes:
             for m in change.changes.modified:
                 m.sbs_chunks = _make_side_by_side(m.chunks)
-    
+
     @with_branch_lock
     def get_filelist(self, inv, file_id, sort_type=None):
         """
@@ -861,25 +861,26 @@
         path = inv.id2path(file_id)
         file_list = []
 
+        revid_set = set()
+
+        for filename, entry in dir_ie.children.iteritems():
+            revid_set.add(entry.revision)
+
+        change_dict = {}
+        for change in self.get_changes(list(revid_set)):
+            change_dict[change.revid] = change
+
         for filename, entry in dir_ie.children.iteritems():
             pathname = filename
             if entry.kind == 'directory':
                 pathname += '/'
 
             revid = entry.revision
-            if self._change_cache:
-                timestamp = self.get_changes([revid])[0].date
-            else:
-                revision = self._branch.repository.get_revision(revid)
-                timestamp = datetime.datetime.fromtimestamp(revision.timestamp)
-
-            change = util.Container(date=timestamp,
-                                    revno=self.get_revno(revid))
 
             file = util.Container(
                 filename=filename, executable=entry.executable, kind=entry.kind,
                 pathname=pathname, file_id=entry.file_id, size=entry.text_size,
-                revid=revid, change=change)
+                revid=revid, change=change_dict[revid])
             file_list.append(file)
 
         if sort_type == 'filename' or sort_type is None:
@@ -904,15 +905,15 @@
         z = time.time()
         lineno = 1
         parity = 0
-        
+
         file_revid = self.get_inventory(revid)[file_id].revision
         oldvalues = None
-        
+
         # because we cache revision metadata ourselves, it's actually much
         # faster to call 'annotate_iter' on the weave directly than it is to
         # ask bzrlib to annotate for us.
         w = self._branch.repository.weave_store.get_weave(file_id, self._branch.repository.get_transaction())
-        
+
         revid_set = set()
         for line_revid, text in w.annotate_iter(file_revid):
             revid_set.add(line_revid)
@@ -923,7 +924,7 @@
                                      change=util.Container())
                 return
         change_cache = dict([(c.revid, c) for c in self.get_changes(list(revid_set))])
-        
+
         last_line_revid = None
         for line_revid, text in w.annotate_iter(file_revid):
             if line_revid == last_line_revid:
@@ -941,11 +942,10 @@
             yield util.Container(parity=parity, lineno=lineno, status=status,
                                  change=change, text=util.fixed_width(text))
             lineno += 1
-        
+
         self.log.debug('annotate: %r secs' % (time.time() - z,))
 
     @with_branch_lock
-    @with_bzrlib_read_lock
     def get_bundle(self, revid, compare_revid=None):
         if compare_revid is None:
             parents = self._revision_graph[revid]

=== added file 'loggerhead/tests/test_filechangecache.py'
--- loggerhead/tests/test_filechangecache.py	1970-01-01 00:00:00 +0000
+++ loggerhead/tests/test_filechangecache.py	2007-09-07 08:09:52 +0000
@@ -0,0 +1,95 @@
+"""Tests for the FileChangeCache in loggerhead.changecache."""
+
+import shutil
+import tempfile
+
+from loggerhead.changecache import FileChangeCache
+
+class MockEntry:
+    def __init__(self, revid):
+        self.revid = revid
+
+
+class MockHistory:
+    def __init__(self):
+        self.fetched_revids = set()
+    def get_file_changes_uncached(self, entries):
+        output = []
+        for entry in entries:
+            self.fetched_revids.add(entry.revid)
+            output.append(entry.revid)
+        return output
+
+
+class TestFileChangeCache(object):
+
+    # setup_method and teardown_method are so i can run the tests with
+    # py.test and take advantage of the error reporting.
+    def setup_method(self, meth):
+        self.setUp()
+
+    def teardown_method(self, meth):
+        self.tearDown()
+
+    def setUp(self):
+        self.cache_folders = []
+
+    def tearDown(self):
+        for folder in self.cache_folders:
+            shutil.rmtree(folder)
+
+
+    def makeHistoryAndEntriesForRevids(self, revids, fill_cache_with=[]):
+        cache_folder = tempfile.mkdtemp()
+        self.cache_folders.append(cache_folder)
+        self.history = MockHistory()
+        self.cache = FileChangeCache(self.history, cache_folder)
+
+        self.entries = [MockEntry(revid) for revid in revids]
+
+        self.cache.get_file_changes([entry for entry in self.entries
+                                     if entry.revid in fill_cache_with])
+        self.history.fetched_revids.clear()
+
+
+    def test_empty_cache(self):
+        """An empty cache passes all the revids through to the history object.
+        """
+        revids = ['a', 'b']
+        self.makeHistoryAndEntriesForRevids(revids)
+
+        result = self.cache.get_file_changes(self.entries)
+
+        assert result == self.revids
+        assert self.history.fetched_revids == set(self.revids)
+
+    def test_full_cache(self):
+        """A full cache passes none of the revids through to the history
+        object.
+        """
+        revids = ['a', 'b']
+        self.makeHistoryAndEntriesForRevids(revids, fill_cache_with=revids)
+
+        result = self.cache.get_file_changes(self.entries)
+
+        assert result == revids
+        assert self.history.fetched_revids == set()
+
+    def test_partial_cache(self):
+        """A partially full cache passes some of the revids through to the
+        history object, and preserves the ordering of the argument list.
+        """
+        # To test the preservation of argument order code, we put the uncached
+        # revid at the beginning, middle and then end of the list of revids
+        # being asked for.
+        for i in range(3):
+            cached_revids = ['a', 'b']
+            revids = cached_revids[:]
+            revids.insert(i, 'uncached')
+            self.makeHistoryAndEntriesForRevids(
+                revids, fillCacheWith=cached_revids)
+
+            result = self.cache.get_file_changes(self.entries)
+            assert result == revids
+
+            assert self.history.fetched_revids == set(['uncached'])

=== modified file 'loggerhead/tests/test_simple.py'
--- loggerhead/tests/test_simple.py	2007-06-26 19:23:08 +0000
+++ loggerhead/tests/test_simple.py	2007-09-07 08:09:52 +0000
@@ -23,9 +23,8 @@
 
 class BasicTests(object):
 
-
-    # there are so i can run it with py.test and take advantage of the
-    # error reporting...
+    # setup_method and teardown_method are so i can run the tests with
+    # py.test and take advantage of the error reporting.
     def setup_method(self, meth):
         self.setUp()
 

